           (*-------------------------------------------------*
            |                 (a part of) ep2                 |
            |                  September 2004                 |
            |                   December 2004 (modified)      |
            |                   November 2005 (modified)      |
            |                      April 2006 (modified)      |
            |                      March 2007  (modified)     |
            |                     August 2007  (modified)     |
            |                                                 |
            |  Markus Roggenbach (Univ of Wales Swansea, UK)  |
            |  Yoshinao Isobe    (AIST, Japan)                |
            *-------------------------------------------------*)

theory ep2_nucleusDF
imports DFP ep2_nucleus
begin

(*********************************************************
                     abstract level
 *********************************************************)

datatype AbsName = Abstract | Loop

consts 
  Absfun :: "AbsName => (AbsName, Event) proc"

primrec
  "Absfun Abstract
          = c !? x -> $Loop"

  "Absfun Loop
          = c !? x -> (SKIP |~| c !? x -> $Loop)"

defs (overloaded)
Set_Absfun_def [simp]: "PNfun == Absfun"

consts 
  Abs :: "(AbsName, Event) proc"

defs
  Abs_def: "Abs == $Abstract"

(*********************************************************
               guard lemmas (routine work)
 *********************************************************)

lemma guardedfun_Abs[simp]:
      "guardedfun Absfun"
by (simp add: guardedfun_def, rule allI, induct_tac p, simp_all)

(*********************************************************
        relating function between ACName and AbsName
 *********************************************************)

consts 
  Abs_to_AC :: "AbsName => (ACName, Event) proc"

primrec
  "Abs_to_AC (Abstract)
          = ($Acquirer |[range c]| $Terminal)"

  "Abs_to_AC (Loop)
          = ($AcConfigManagement |[range c]| $TerminalConfigManagement)"

(*********************************************************
           a theorem for verifying Abs <=F AC
 *********************************************************)

theorem ep2_Abs_AC: "Abs <=F AC"
apply (unfold Abs_def AC_def)
apply (rule cspF_fp_induct_left[of _ "Abs_to_AC"], auto)
apply (induct_tac p, auto)

apply ((tactic {* cspF_hsf_tac 1 *})+, 
       (rule | rule cspF_decompo_ref |
        rule cspF_Int_choice_left1, auto |
        rule cspF_Int_choice_left2, rule cspF_decompo_ref, auto)?,
       (auto simp add: image_iff inj_on_def)?)+
done

(*********************************************************
        relating function between ACName and AbsName
 *********************************************************)

consts 
  Abs_to_DF :: "AbsName => (DFtickName, Event) proc"

primrec
  "Abs_to_DF (Abstract) = ($DFtick)"
  "Abs_to_DF (Loop) = ($DFtick)"

(*********************************************************
           a theorem for verifying Abs <=F AC
 *********************************************************)

theorem ep2_DF_Abs: "$DFtick <=F Abs"
apply (unfold Abs_def)
apply (rule cspF_fp_induct_right[of _ _ "Abs_to_DF"], auto)
apply (induct_tac p, auto)
apply ((tactic {* cspF_hsf_tac 1 *}, rule cspF_Int_choice_left1, rule cspF_decompo_ref, auto) |
       (tactic {* cspF_hsf_tac 1 *}, rule cspF_Int_choice_left2, simp))+
done

(*-------------------------------------------------------*
 |                 AC is Deadlock-free.                  |
 *-------------------------------------------------------*)

theorem AC_isDeadlockFree: "AC isDeadlockFree"
apply (simp add: DeadlockFree_DFtick_ref)
apply (rule cspF_trans[of _ _ _ Abs])
apply (rule ep2_DF_Abs, rule ep2_Abs_AC)
done

end
